from django.shortcuts import render_to_response
from django.http import HttpResponse, HttpResponseRedirect
from django.contrib.auth.decorators import login_required
from django.template import RequestContext

import project.settings

#--------------------------------------------------------------------------------------------------

def user_has_rights(request):
    return request.user.is_authenticated() and request.user.permissions.is_server_admin

#--------------------------------------------------------------------------------------------------

@login_required
def index(request):

    if not user_has_rights(request):
        return HttpResponseRedirect("/")

    import os

    # get contents of log directory
    path = project.settings.CYBRO_SCGI_SERVER_LOG_PATH
    f_list = [f for f in os.listdir(path)]
    f_list.sort()

    # filter files only
    log_files = []
    for f in f_list:
        if not os.path.isdir(path + f):
            size = "%.1f kB" % (os.path.getsize(path + f) / 1024.0)
            log_files.append(
                {
                    "name": f,
                    "path:": path,
                    "size": size,
                }
            )

    return render_to_response("server_admin/index.html",
        {
            "active_menu": "server_admin",
            "log_files": log_files,
        }, context_instance = RequestContext(request)
    )

#--------------------------------------------------------------------------------------------------

@login_required
def edit_config(request, cfg_file):

    if not user_has_rights(request):
        return HttpResponseRedirect("/")

    import project.settings
    err = False
    msg = ""
    config = ""

    if cfg_file == "1":
        fname = project.settings.CYBRO_SCGI_SERVER_CONFIG
        caption = "Cybro SCGI server configuration"
    elif cfg_file == "2":
        fname = project.settings.DATALOGGER_CONFIG
        caption = "Data logger configuration"
    else:
        return HttpResponseRedirect("/")

    if request.method == "POST":
        config = request.POST.get("config")
        try:
            f = open(fname, 'w')
            f.write(config.encode("utf-8"))
            f.close()
            return HttpResponseRedirect("/server_admin/")
        except:
            err = True
            msg = "Error writing file '%s'." % fname
    else:
        try:
            f = open(fname, 'r')
            config = f.read()
            f.close()
        except:
            err = True
            msg = "Error reading file '%s'." % fname

    return render_to_response("server_admin/edit_cfg.html",
        {
            "cfg_file": cfg_file,
            "caption": caption,
            "config": config,
            "err": err,
            "msg": msg,
            "active_menu": "server_admin",
        }, context_instance = RequestContext(request)
    )

#--------------------------------------------------------------------------------------------------

@login_required
def dump_db(request, what):

    if not user_has_rights(request):
        return HttpResponseRedirect("/")

    import project.settings
    import os
    import tempfile
    import gzip
    from django.core.servers.basehttp import FileWrapper

    # define file names
    sql_dump_fname = tempfile.mktemp(prefix = "solardump_", suffix = ".sql")
    gzip_sql_dump_fname = sql_dump_fname + ".gz"

    if what == "pages":
        tab_opt = "--ignore-table=%s.alarms --ignore-table=%s.measurements" % (project.settings.DATABASE_NAME, project.settings.DATABASE_NAME)
        out_fname = "solar_db_dump_pages.sql.gz"
    else:
        tab_opt = "--tables alarms measurements"
        out_fname = "solar_db_dump_measurements.sql.gz"

    # execute mysqldump
    cmd = "mysqldump --add-drop-table -c -u %s -p%s %s %s > %s" % \
        (project.settings.DATABASE_USER, project.settings.DATABASE_PASSWORD, project.settings.DATABASE_NAME, tab_opt, sql_dump_fname)

    os.system(cmd)

    # gzip file
    f_in = open(sql_dump_fname, 'rb')
    f_out = gzip.open(gzip_sql_dump_fname, 'wb')
    f_out.writelines(f_in)
    f_out.close()
    f_in.close()

    os.remove(sql_dump_fname)

    f_in = open(gzip_sql_dump_fname, 'rb')

    response = HttpResponse(FileWrapper(f_in), content_type='application/zip')
    response['Content-Disposition'] = "attachment; filename=%s" % (out_fname)
    return response

#--------------------------------------------------------------------------------------------------

@login_required
def view_log_file(request, log_file):

    if not user_has_rights(request):
        return HttpResponseRedirect("/")

    import os
    path = project.settings.CYBRO_SCGI_SERVER_LOG_PATH
    fname = path + log_file

    err = True
    if os.path.exists(fname):
        try:
            f = open(fname, 'r')
            content = f.read()
            f.close()
            err = False
        except:
            pass

    if err:
        content = "Error reading file '%s'." % fname

    response = HttpResponse(content, content_type='text/plain')
    response['Content-Disposition'] = "attachment; filename=%s" % (log_file)
    return response

#--------------------------------------------------------------------------------------------------

@login_required
def view_db_info(request):

    if not user_has_rights(request):
        return HttpResponseRedirect("/")

    from django.db import connection
    import itertools

    def exec_sql(sql):
        res = []
        cursor = connection.cursor()
        cursor.execute(sql)
        col_names = [desc[0] for desc in cursor.description]

        while True:
            row = cursor.fetchone()
            if row != None:
                res.append(dict(itertools.izip(col_names, row)))
            else:
                break

        return res

    table_info = exec_sql("SHOW TABLE STATUS FROM `%s`" % project.settings.DATABASE_NAME);

    total_data_len = 0
    total_index_len = 0
    for table in table_info:
        total_data_len += table["Data_length"]
        total_index_len += table["Index_length"]

    return render_to_response("server_admin/view_db_info.html",
        {
            "settings": project.settings,
            "table_info": table_info,
            "total_data_len": total_data_len,
            "total_index_len": total_index_len,
            "total_len": total_data_len + total_index_len,
            "active_menu": "server_admin",
        }, context_instance = RequestContext(request)
    )

#--------------------------------------------------------------------------------------------------

@login_required
def restart_scgi_server(request):

    if not user_has_rights(request):
        return HttpResponseRedirect("/")

    from project.main.common import create_comm_request
    from django.http import QueryDict
    request.GET = QueryDict("sys.server_shutdown=1")
    create_comm_request(request)

    return HttpResponseRedirect("/server_admin/")

#--------------------------------------------------------------------------------------------------

def test_timeout(request):
    return render_to_response("server_admin/test_timeout.html",
        {
        }, context_instance = RequestContext(request)
    )

#--------------------------------------------------------------------------------------------------

def test_timeout_get_data(request):
    import time
    time.sleep(3)

    return HttpResponse("ok")

#--------------------------------------------------------------------------------------------------
